#ifndef RANDOMMANAGER_H
#define RANDOMMANAGER_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>

#include "logger.h"

using std::vector;

class randommanager;

class randomvariable {
public:
  randomvariable(int _n_events, int _switch_IS_MC, int _n_steps, int _n_bins, const std::string &_name);
  void optimize(const double psp_weight);
  int do_optimization_step(int events);
  void get_random(double &x, double &g_IS);
  void get_g_IS(double r, double &g_IS);
  void multiply_g_IS(double r, double & g_IS);
  double calc_variation();
  void proceeding_out(std::ofstream &out_proceeding);
  void proceeding_in(int &proc, vector<std::string> &readin);
  void check_proceeding_in(int &int_end, int &proc, vector<std::string> &readin);
  void proceeding_size(int &proc);
  void proceeding_size(vector<int> & save_proceeding);
  void proceeding_save();
  void save_weights(std::ofstream &out_weights);
  int readin_weights(std::vector<std::string> &readin);

  int channel;

  int n_events;
  int switch_IS_MC;
  int n_steps;
  int n_bins;

  int opt_end;
  bool used;

  bool imp_sampling;
  
  vector<double> s;
  vector<double> save_s;

  std::vector<double> alpha;
  vector<double> beta;
  vector<int> cuts_channel;
  vector<int> counts_channel;
  vector<double> sum_channel_weight;
  vector<double> sum_channel_weight2;

  std::string name;
  std::string filename;

  randommanager *manager;

private:
  void init(int _n_events, int _switch_IS_MC, int _n_steps, int _n_bins, const std::string & _name);
  void get_random(double &x, double &g_IS, vector<double> &s);

  int save_opt_end;

  vector<double> save_alpha;
  vector<double> save_beta;
  vector<int> save_counts_channel;
  vector<int> save_cuts_channel;
  vector<double> save_sum_channel_weight;
  vector<double> save_sum_channel_weight2;
};


class randommanager{
public:
  randommanager();
  randommanager(vector<double> sran, std::string _weightdir, std::string _processname);
  ~randommanager();

  void register_variable(randomvariable* new_var, bool imp_sampling);
  void optimize(const double psp_weight, const double psp_weight2);
  void increase_cut_counter();
  void do_optimization_step(int events);
  void proceeding_out(std::ofstream &out_proceeding);
  void proceeding_in(int &proc, vector<std::string> &readin);
  void check_proceeding_in(int &int_end, int &proc, vector<std::string> &readin);
  void proceeding_size(int &proc);
  void proceeding_size(vector<int> & save_proceeding);
  void proceeding_save();
  void writeout_weights();

  void add_var_to_queue(randomvariable* variable);
  //  void nancut(bool count_as_cut);
  void nancut();

private:
  void readin_weights();

  vector<randomvariable*> random_variables;

  bool opt_end;
  vector<double> s;
  Logger logger;

  std::string weightdir;
  std::string processname;

  vector<double> save_s;

  vector<randomvariable*> used_queue;
  int queue_counter;
  int queue_threshold;

  vector<std::string> readin;
};

#endif
