#include "../include/classes.cxx"

void observable_set::initialization_distribution(){
  Logger logger("observable_set::initialization_distribution");
  logger << LOG_DEBUG << "started" << endl;

// **************************************************************************
// *                                                                        *
// *  reading parameters from 'file_distribution.dat'                       *
// *                                                                        *
// **************************************************************************
  logger << LOG_DEBUG << "switch_distribution = " << switch_distribution << endl;
  if (switch_distribution){readin_file_distribution();}
  logger << LOG_INFO << "List of differential distributions:" << endl << endl;
  for (int i_d = 0; i_d < dat.size(); i_d++){logger << LOG_INFO << "No. " << setw(3) << i_d << ":" << dat[i_d] << endl;}
  logger.newLine(LOG_INFO);


// **************************************************************************
// *                                                                        *
// *  reading parameters from 'file_dddistribution.dat'                     *
// *                                                                        *
// **************************************************************************
  if (switch_distribution){readin_file_dddistribution();}
  logger << LOG_INFO << "List of doubly-differential distributions:" << endl << endl;
  for (int i_ddd = 0; i_ddd < dddat.size(); i_ddd++){logger << LOG_INFO << "dddistribution " << setw(3) << i_ddd << ":" << dddat[i_ddd] << endl;}
  logger.newLine(LOG_INFO);



  int isystem = 0; 
  string xorder;
  string temp_directory;

  string dir_distribution = "distribution";
  if (switch_output_distribution){system_execute(logger, "mkdir " + dir_distribution, isystem);}
  filename_distribution.resize(dat.size());
  filename_dddistribution.resize(dddat.size());
  for (int i = 0; i < dat.size(); i++){
    filename_distribution[i] = dir_distribution + "/" + dat[i].xdistribution_name + "_" + name_process + ".dat";
  }
  for (int i = 0; i < dddat.size(); i++){
    filename_dddistribution[i] = dir_distribution + "/" + dddat[i].name + "_" + name_process + ".dat";
  }
  
  if (switch_CV){
    if (switch_output_distribution){system_execute(logger, "mkdir " + dir_distribution + "/CV", isystem);}
    filename_distribution_CV.resize(n_scales_CV, filename_distribution);
    filename_dddistribution_CV.resize(n_scales_CV, filename_dddistribution);

    logger << LOG_DEBUG_VERBOSE << "directory_name_scale_CV.size() = " << directory_name_scale_CV.size() << endl;
    for (int i_s = 0; i_s < n_scales_CV; i_s++){
      logger << LOG_DEBUG_VERBOSE << "directory_name_scale_CV[" << i_s << "] = " << directory_name_scale_CV[i_s] << endl;
      if (switch_output_distribution){system_execute(logger, "mkdir " + dir_distribution + "/CV/" + directory_name_scale_CV[i_s], isystem);}
      logger << LOG_DEBUG_VERBOSE << xorder << "   execution status: " << isystem << "." << endl;
    
      filename_distribution_all_CV = dir_distribution + "/CV" + "/distribution_" + name_process + ".txt";
      // Should replace the following files at some point...

      for (int i_d = 0; i_d < dat.size(); i_d++){
	filename_distribution_CV[i_s][i_d]  = dir_distribution + "/CV/" + directory_name_scale_CV[i_s] + "/" + dat[i_d].xdistribution_name + "_" + name_process + ".dat";
      }
      for (int i_d = 0; i_d < dddat.size(); i_d++){
	filename_dddistribution_CV[i_s][i_d]  = dir_distribution + "/CV/" + directory_name_scale_CV[i_s] + "/" + dddat[i_d].name + "_" + name_process + ".dat";
      }
    }
  }

  if (switch_TSV){
    logger << LOG_DEBUG_VERBOSE << "before bin_count_TSV" << endl;
    bin_count_TSV.resize(n_qTcut, vector<vector<double> > (dat.size() + dddat.size()));
    for (int i_q = 0; i_q < n_qTcut; i_q++){
      for (int i_d = 0; i_d < dat.size(); i_d++){
	bin_count_TSV[i_q][i_d].resize(dat[i_d].n_bins, 0);
      }
      for (int i_d = 0; i_d < dddat.size(); i_d++){
	int i_ddd = dat.size() + i_d;
	bin_count_TSV[i_q][i_ddd].resize(dddat[i_d].n_bins, 0);
      }
    }

    logger << LOG_DEBUG_VERBOSE << "before bin_weight_TSV" << endl;

    // n_qTcut will typically be by far too large
    // define a selection of a handful of qTcut values instead and use their number n_qTcut_distribution (or so...) !!!

    bin_weight_TSV.resize(n_set_TSV, vector<vector<vector<vector<vector<double> > > > > (n_qTcut));
    bin_weight2_TSV.resize(n_set_TSV, vector<vector<vector<vector<vector<double> > > > > (n_qTcut));
    for (int i_s = 0; i_s < n_set_TSV; i_s++){
      for (int i_q = 0; i_q < n_qTcut; i_q++){
	bin_weight_TSV[i_s][i_q].resize(n_scale_ren_TSV[i_s], vector<vector<vector<double> > > (n_scale_fact_TSV[i_s], vector<vector<double> > (dat.size() + dddat.size())));
	bin_weight2_TSV[i_s][i_q].resize(n_scale_ren_TSV[i_s], vector<vector<vector<double> > > (n_scale_fact_TSV[i_s], vector<vector<double> > (dat.size() + dddat.size())));
	for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	  for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	    for (int i_d = 0; i_d < dat.size(); i_d++){
	      bin_weight_TSV[i_s][i_q][i_r][i_f][i_d].resize(dat[i_d].n_bins, 0.);
	      bin_weight2_TSV[i_s][i_q][i_r][i_f][i_d].resize(dat[i_d].n_bins, 0.);
	    }
	    for (int i_d = 0; i_d < dddat.size(); i_d++){
	      int i_ddd = dat.size() + i_d;
	      bin_weight_TSV[i_s][i_q][i_r][i_f][i_ddd].resize(dddat[i_d].n_bins, 0.);
	      bin_weight2_TSV[i_s][i_q][i_r][i_f][i_ddd].resize(dddat[i_d].n_bins, 0.);
	    }
	  }
	}
      }
    }
    
    logger << LOG_DEBUG_VERBOSE << "after bin_weight_TSV" << endl;

    filename_distribution_TSV.resize(n_set_TSV);
    for (int i_s = 0; i_s < n_set_TSV; i_s++){
      string temp_directory = dir_distribution + "/" + name_set_TSV[i_s];
      if (switch_output_distribution){system_execute(logger, "mkdir " + temp_directory, isystem);}
      filename_distribution_TSV[i_s] = temp_directory + "/distribution_" + name_process + ".txt";
    }
  }




  logger << LOG_DEBUG << "finished" << endl;
}

void observable_set::readin_file_distribution(){
  Logger logger("observable_set::readin_file_distribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

// **************************************************************************
// *                                                                        *
// *  reading parameters from 'file_distribution.dat'                       *
// *                                                                        *
// **************************************************************************

  char LineBuffer[512];
  string s0;
  //  int counter = 0;
  string readindata;
  vector<string> readin, vs0;
  vector<string> user_variable;
  vector<string> user_variable_counter;
  vector<string> user_value;
  string pathname = "" + path_to_main + "/file_distribution.dat";
  logger << LOG_DEBUG << "path_to_main = " << path_to_main << endl;
  logger << LOG_DEBUG << "pathname = " << pathname << endl;
  ifstream in_new_distribution_gen(pathname.c_str());
  while (in_new_distribution_gen.getline(LineBuffer, 512)){readin.push_back(LineBuffer);}
  in_new_distribution_gen.close();
  
  if (readin.size() == 0) {
    logger << LOG_INFO << "file_distributions.dat could not be read in" << endl;
    //    assert(readin.size() > 0);
  }

  vector<string> user_variable_additional;
  get_userinput_extravalue_from_readin(user_variable, user_variable_additional, user_value, readin);

  vector<string> distributionname;
  vector<string> distributiontype;
  
  vector<vector<vector<int> > > all_particle;
  vector<vector<vector<int> > > required_subparticle;
  vector<vector<vector<int> > > all_particle_group;
  vector<vector<vector<int> > > all_particle_order;

  vector<double> startpoint;
  vector<double> endpoint;
  vector<double> binwidth;
  vector<string> edges;
  vector<int> binnumber;
  vector<string> binningtype;
  int current_counter = 0;
  vector<int> vi0(0);
  vector<vector<int> > vvi0(0);
  vector<vector<int> > vvi3(3);
  current_counter = -1;

  for (int i = 0; i < user_variable.size(); i++){
    if (user_variable[i] == "distributionname"){
      current_counter++;
      distributionname.push_back(s0);
      distributiontype.push_back(s0);
      all_particle.push_back(vvi0);
      required_subparticle.push_back(vvi0);
      all_particle_group.push_back(vvi0);
      all_particle_order.push_back(vvi0);
      startpoint.push_back(0.);
      endpoint.push_back(0.);
      binwidth.push_back(0.);
      edges.push_back("");
      binnumber.push_back(0);
      binningtype.push_back("linear");

      distributionname[current_counter] = user_value[i];
    }
    else if (user_variable[i] == "distributiontype"){distributiontype[current_counter] = user_value[i];}
    else if (user_variable[i] == "startpoint"){startpoint[current_counter] = atof(user_value[i].c_str());}
    else if (user_variable[i] == "endpoint"){endpoint[current_counter] = atof(user_value[i].c_str());}
    else if (user_variable[i] == "binnumber"){binnumber[current_counter] = atoi(user_value[i].c_str());}
    else if (user_variable[i] == "binwidth"){binwidth[current_counter] = atof(user_value[i].c_str());}
    else if (user_variable[i] == "edges"){edges[current_counter] = user_value[i];}
    else if (user_variable[i] == "binningtype"){binningtype[current_counter] = user_value[i].c_str();}
    else if (user_variable[i] == "particle"){
      //      logger << LOG_DEBUG << "new: distribution " << current_counter << endl;
      //      logger << LOG_DEBUG << "user_variable[" << setw(3) << i << "] = " << user_variable[i] << "   user_variable_additional[" << setw(3) << i << "] = " << user_variable_additional[i] << "   user_value[" << setw(3) << i << "] = " << user_value[i] << endl;
      int temp_i = 1;
      int temp_r = 2; // former default: 1 !!!
      if (user_variable_additional[i] != ""){
	vector<string> temp_s(1);
	for (int i_s = 0; i_s < user_variable_additional[i].size(); i_s++){
	  if (user_variable_additional[i][i_s] == ' ' || user_variable_additional[i][i_s] == char(9)){
	    if (temp_s[temp_s.size() - 1].size() == 0){}
	    else {temp_s.push_back("");}
	  }
	  else {temp_s[temp_s.size() - 1].push_back(user_variable_additional[i][i_s]);}
	}	
	if (temp_s.size() > 0){
	  temp_i = atoi(temp_s[0].c_str());
	  if (temp_s.size() == 2){
	    if (temp_s[1] == "+"){temp_r = 1;}
	    else if (temp_s[1] == "x"){temp_r = 2;}
	    else if (temp_s[1] == "0"){temp_r = 0;}
	    else if (temp_s[1] == "-"){temp_r = -1;}
	    else {logger << LOG_ERROR<< "entry '" << temp_s[1] << "' not allowed!" << endl; assert(false); exit(1);}
	  }
	}
	// before introduction of '+' and '-'
	//	temp_i = atoi(user_variable_additional[i].c_str());
      }
      if (temp_i > all_particle[current_counter].size()){
	all_particle[current_counter].resize(temp_i);
	all_particle_group[current_counter].resize(temp_i);
	required_subparticle[current_counter].resize(temp_i);
      }
      required_subparticle[current_counter][temp_i - 1].push_back(temp_r);

      vector<string> temp_s(1);
      for (int i_s = 0; i_s < user_value[i].size(); i_s++){
	if (user_value[i][i_s] == ' ' || user_value[i][i_s] == char(9)){
	  if (temp_s[temp_s.size() - 1].size() == 0){}
	  else{temp_s.push_back("");}
	}
	else {temp_s[temp_s.size() - 1].push_back(user_value[i][i_s]);}
      }
      //      logger << LOG_DEBUG << "current_counter = " << current_counter << endl;
      //      logger << LOG_DEBUG << "temp_i = " << temp_i << endl;
      //      logger << LOG_DEBUG << "temp_s.size() = " << temp_s.size() << endl;
      //      for (int i_t = 0; i_t < temp_s.size(); i_t++){logger << LOG_DEBUG << "temp_s[" << i_t << "] = " << temp_s[i_t] << endl;}
      if (temp_s.size() == 1){
	all_particle_group[current_counter][temp_i - 1].push_back(0);
	all_particle[current_counter][temp_i - 1].push_back(atoi(temp_s[0].c_str()));
      }
      else if (temp_s.size() == 2){
	if (esi.observed_object[temp_s[0]] != 0 || temp_s[0] == "all"){
	  //	  all_particle_group[current_counter][temp_i - 1].push_back(esi.observed_object[temp_s[0]]);
	  all_particle_group[current_counter][temp_i - 1].push_back(esi.observed_object[equivalent_object[temp_s[0]]]);

	  //	  logger << LOG_DEBUG << "new: distribution no. " << current_counter << "   temp_s[0] = " << setw(7) << temp_s[0] << "   esi.observed_object[" << setw(7) << temp_s[0] << "] = " << setw(7) << esi.observed_object[temp_s[0]] << "   no_relevant_object[equivalent_object[" << setw(7) << temp_s[0] << "] = " << setw(7) << equivalent_object[temp_s[0]] << "] = " << no_relevant_object[equivalent_object[temp_s[0]]] << endl;

	  //	  logger << LOG_DEBUG << "new: distribution no. " << current_counter << "   temp_s[0] = " << setw(7) << temp_s[0] << "   esi.observed_object[" << setw(7) << temp_s[0] << "] = " << setw(7) << esi.observed_object[temp_s[0]] << "   esi.observed_object[equivalent_object[" << setw(7) << temp_s[0] << "] = " << setw(7) << equivalent_object[temp_s[0]] << "] = " << setw(7) << esi.observed_object[equivalent_object[temp_s[0]]] << endl;
	  //equivalent_object
	  //equivalent_object, map<int, int> & equivalent_no_object, map<string, int> & no_relevant_object
	  all_particle[current_counter][temp_i - 1].push_back(atoi(temp_s[1].c_str()));
	}
	else {
          logger << LOG_ERROR << "object not known: " << temp_s[0] << endl;
          assert(false);
	  exit (1);
	}
      }
      else {
        logger << LOG_ERROR << "Input not understood!" << endl;
        logger << LOG_ERROR << "user_variable[" << setw(3) << i << "] = " << user_variable[i] << "   user_variable_additional[" << setw(3) << i << "] = " << user_variable_additional[i] << "   user_value[" << setw(3) << i << "] = " << user_value[i] << endl;
        assert(false);
	exit(1);
      }
    }
  }
  
  for (int i_d = 0; i_d < current_counter + 1; i_d++){
    logger << LOG_DEBUG << "distributionname[" << i_d << "] = " << distributionname[i_d] << endl;
    xdistribution one_dist(distributionname[i_d], distributiontype[i_d], all_particle[i_d], all_particle_group[i_d], required_subparticle[i_d], startpoint[i_d], endpoint[i_d], binnumber[i_d], binwidth[i_d], edges[i_d], binningtype[i_d], this); 
    dat.push_back(one_dist);
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


//vector<dddistribution> & all_dddistributions, vector<xdistribution> & all_distributions, observable_set & oset
void observable_set::readin_file_dddistribution(){
  Logger logger("observable_set::readin_file_dddistributions");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

// **************************************************************************
// *                                                                        *
// *  reading parameters from 'file_distribution.dat'                       *
// *                                                                        *
// **************************************************************************
  char LineBuffer[256];
  string s0;
  //  int counter = 0;
  string readindata;
  vector<string> readin, vs0;
  vector<string> user_variable;
  vector<string> user_variable_counter;
  vector<string> user_value;
  string pathname = "" + path_to_main + "/file_dddistribution.dat";
  logger << LOG_DEBUG << "pathname = " << pathname << endl;

  ifstream in_new_distribution_gen(pathname.c_str());
  while (in_new_distribution_gen.getline(LineBuffer, 256)) {
    readin.push_back(LineBuffer);
  }
  in_new_distribution_gen.close();

  for (int i = 0; i < readin.size(); i++){
    readindata = readin[i][0];
    if (readindata != "/" && readindata != "#" && readindata != "%"){
      int start = 0;
      user_variable.push_back(s0);
      user_variable_counter.push_back(s0);
      user_value.push_back(s0);
      
      for (int j = 0; j < readin[i].size(); j++){
	if (start == 0 || start == 1){
	  if (((readin[i][j] == ' ') || (readin[i][j] == char(9))) && start == 0){}
	  else if ((readin[i][j] != ' ') && (readin[i][j] != char(9))){
	    user_variable[user_variable.size() - 1].push_back(readin[i][j]);
	    if (start != 1){start = 1;}
	  }
	  else {start++;}
	}
	else if (start == 2){
	  if (readin[i][j] == '='){start++;}
	  else if ((readin[i][j] == ' ') || (readin[i][j] == char(9))){}
	  else {logger << LOG_DEBUG << "Incorrect input in line " << i << endl;
	    user_variable.erase(user_variable.end(), user_variable.end());
	    user_value.erase(user_value.end(), user_value.end());
	    break;
	  }
	}
	else if (start == 3 || start == 4){
	  if (((readin[i][j] == ' ') || (readin[i][j] == char(9))) && start == 3){}
	  else if ((readin[i][j] != ' ') && (readin[i][j] != char(9))){
	    if (readin[i][j] != char(13)){user_value[user_value.size() - 1].push_back(readin[i][j]);}
	    if (start != 4){start = 4;}
	  }
	  else {start++;}
	}
	else {break;}
      }
    }
  }


  vector<string> dddistributionname;
  vector<string> distributionname_1;
  vector<string> distributionname_2;
  vector<int> d_1;
  vector<int> d_2;
  vector<int> n_bins;
  int current_counter = 0;
  vector<int> vi0(0);
  current_counter = -1;
  for (int i = 0; i < user_variable.size(); i++){
      logger << LOG_DEBUG << "user_variable[" << i << "] = " << user_variable[i] << "   user_value[" << i << "] = " << user_value[i] << endl;
    if (user_variable[i] == "dddistributionname"){
      current_counter++;
      dddistributionname.push_back(s0);
      distributionname_1.push_back(s0);
      distributionname_2.push_back(s0);
      //      d_1.push_back(0);
      //      d_2.push_back(0);
      //      n_bins.push_back(0);
      dddistributionname[current_counter] = user_value[i];
    }
    else if (user_variable[i] == "distributionname_1"){distributionname_1[current_counter] = user_value[i];}
    else if (user_variable[i] == "distributionname_2"){distributionname_2[current_counter] = user_value[i];}
  }
    
  for (int d = 0; d < current_counter + 1; d++){
      logger << LOG_DEBUG << "dddistributionname[" << d << "] = " << dddistributionname[d] << endl;
    int d_1 = -1;
    int d_2 = -1;
    //    int n_bins;
    for (int id = 0; id < dat.size(); id++){
      if (dat[id].xdistribution_name == distributionname_1[d]){d_1 = id;}
      if (dat[id].xdistribution_name == distributionname_2[d]){d_2 = id;}
    }
    if (d_1 != -1 && d_2 != -1){
      //      n_bins = dat[d_1].n_bins * dat[d_2].n_bins;
      logger << LOG_DEBUG << "dddistributionname[" << d << "] = " << dddistributionname[d] << "   d_1 = " << d_1 << "   d_2 = " << d_2 << endl;
      dddistribution one_dddist(dddistributionname[d], d_1, d_2, dat);
      //      dddistribution one_dddist(dddistributionname[d], d_1, d_2, dat[d_1], dat[d_2]);
      dddat.push_back(one_dddist);
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}



void observable_set::determine_extended_distribution(){//, observable_set & oset
  Logger logger("determine_extended_distribution");
  logger << LOG_DEBUG << "called" << endl;
  //  static int initialization = 1;

  //  int plotmode = 1;
  //  int n_subgroup = subgroup.size();
  // !!! subgroup not needed !!!
  string vs;
  string s0;
  vector<string> vs0;
  string filename;
  vector<string> readin;
  //  char LineBuffer[128];

  //  int QCD_order = 0;

  //#include "../include/declarations.cxx"
  /*
  map<string, int> observed_object;
  vector<string> object_list;
  object_list.reserve(30);
  vector<int> object_category;
  object_category.reserve(30);
  //  vector<string> relevant_object_list;
  //  vector<int> relevant_object_category;

  temp_define_object_list(observed_object, object_list, object_category);

  assert(esi.observed_object == observed_object);
  assert(esi.observed_list == observed_list);
  assert(esi.observed_category == observed_category);


  //  map<string, string> equivalent_object;
  //  map<int, int> equivalent_no_object;
  //  map<string, int> no_relevant_object;
  */

// **************************************************************************
// *                                                                        *
// *  reading parameters from 'file_distribution.dat'                       *
// *                                                                        *
// **************************************************************************
//  vector<xdistribution> all_distribution;
//  simplified_readin_distributions(all_distribution, oset);
  readin_file_distribution();
  vector<xdistribution> all_distribution = dat;
  logger << LOG_DEBUG << "all_distribution.size() = " << all_distribution.size() << endl;
  for (int d = 0; d < all_distribution.size(); d++){logger << LOG_DEBUG << "distribution[" << d << "] = " << endl << all_distribution[d] << endl;}

// **************************************************************************
// *                                                                        *
// *  reading parameters from 'file_dddistribution.dat'                     *
// *                                                                        *
// **************************************************************************
//  vector<dddistribution> all_dddistribution;
//  simplified_readin_dddistributions(all_dddistribution, all_distribution, oset);
  readin_file_dddistribution();
  vector<dddistribution> all_dddistribution = dddat;

  logger << LOG_DEBUG << "all_dddistribution.size() = " << all_dddistribution.size() << endl;
  for (int d = 0; d < all_dddistribution.size(); d++){logger << LOG_DEBUG << "all_dddistribution[" << d << "] = " << endl << all_dddistribution[d] << endl;}
 
  vector<int> mapping_dddistribution(all_distribution.size(), 0);
  // should be available for all subroutines !!!

  //  vector<xdistribution> extended_distribution = all_distribution;
  extended_distribution = all_distribution;
  extended_distribution.reserve(all_distribution.size() + all_dddistribution.size() * 2);
  //  extended_distribution.reserve(all_distribution.size() + all_dddistribution.size());
    logger << LOG_DEBUG << "extended_distribution.size() = " << extended_distribution.size() << endl;

  for (int i_d = 0; i_d < all_dddistribution.size(); i_d++){
    vector<int> vi;
    string type;
    //    double asym_n_bins;
    if ((((all_dddistribution[i_d].distribution_1).xdistribution_type == "y" ||
	  (all_dddistribution[i_d].distribution_1).xdistribution_type == "absydiff" ||
	  (all_dddistribution[i_d].distribution_1).xdistribution_type == "eta" ||
	  (all_dddistribution[i_d].distribution_1).xdistribution_type == "absetadiff") &&
	 ((all_dddistribution[i_d].distribution_1).n_bins == 2)) &&
	(((all_dddistribution[i_d].distribution_2).xdistribution_type == "y" ||
	  (all_dddistribution[i_d].distribution_2).xdistribution_type == "absydiff" ||
	  (all_dddistribution[i_d].distribution_2).xdistribution_type == "eta" ||
	  (all_dddistribution[i_d].distribution_2).xdistribution_type == "absetadiff") &&
	 ((all_dddistribution[i_d].distribution_2).n_bins == 2))){
      type = "doubleasym";
    }
    else if ((((all_dddistribution[i_d].distribution_1).xdistribution_type == "y" ||
	       (all_dddistribution[i_d].distribution_1).xdistribution_type == "absydiff" ||
	       (all_dddistribution[i_d].distribution_1).xdistribution_type == "eta" ||
	       (all_dddistribution[i_d].distribution_1).xdistribution_type == "absetadiff")) &&
	     ((all_dddistribution[i_d].distribution_1).n_bins == 2)){
      xdistribution temp(all_dddistribution[i_d].name, "asym1", (all_dddistribution[i_d].distribution_2).all_particle, (all_dddistribution[i_d].distribution_2).all_particle_group, (all_dddistribution[i_d].distribution_2).required_subparticle, all_dddistribution[i_d].distribution_2.start, 0, all_dddistribution[i_d].distribution_2.n_bins, all_dddistribution[i_d].distribution_2.step, "", "linear");//, 0, 3);
      extended_distribution.push_back(temp);
      mapping_dddistribution.push_back(i_d);
      xdistribution tempsum("sum" + all_dddistribution[i_d].name, "sumasym1", (all_dddistribution[i_d].distribution_2).all_particle, (all_dddistribution[i_d].distribution_2).all_particle_group, (all_dddistribution[i_d].distribution_2).required_subparticle, all_dddistribution[i_d].distribution_2.start, 0, all_dddistribution[i_d].distribution_2.n_bins, all_dddistribution[i_d].distribution_2.step, "", "linear");//, 0, 3);
      extended_distribution.push_back(tempsum);
      mapping_dddistribution.push_back(i_d);
    }
    else if ((((all_dddistribution[i_d].distribution_2).xdistribution_type == "y" ||
	       (all_dddistribution[i_d].distribution_2).xdistribution_type == "absydiff" ||
	       (all_dddistribution[i_d].distribution_2).xdistribution_type == "eta" ||
	       (all_dddistribution[i_d].distribution_2).xdistribution_type == "absetadiff")) &&
	     ((all_dddistribution[i_d].distribution_2).n_bins == 2)){
      xdistribution temp(all_dddistribution[i_d].name, "asym2", (all_dddistribution[i_d].distribution_1).all_particle, (all_dddistribution[i_d].distribution_1).all_particle_group, (all_dddistribution[i_d].distribution_1).required_subparticle, all_dddistribution[i_d].distribution_1.start, 0, all_dddistribution[i_d].distribution_1.n_bins, all_dddistribution[i_d].distribution_1.step, "", "linear");//, 0, 3);
      extended_distribution.push_back(temp);
      mapping_dddistribution.push_back(i_d);
      xdistribution tempsum("sum" + all_dddistribution[i_d].name, "sumasym2", (all_dddistribution[i_d].distribution_1).all_particle, (all_dddistribution[i_d].distribution_1).all_particle_group, (all_dddistribution[i_d].distribution_1).required_subparticle, all_dddistribution[i_d].distribution_1.start, 0, all_dddistribution[i_d].distribution_1.n_bins, all_dddistribution[i_d].distribution_1.step, "", "linear");//, 0, 3);
      extended_distribution.push_back(tempsum);
      mapping_dddistribution.push_back(i_d);
    }
    else {
      type = "dddistribution";
      cout << "Seems like double-distributions have not been properly installed yet!" << endl;

      xdistribution temp = get_fake_distribution_from_dddistribution(all_dddistribution[i_d]);
      //(all_dddistribution[i_d].name, type, (all_dddistribution[i_d].distribution_1).all_particle, (all_dddistribution[i_d].distribution_1).all_particle_group, (all_dddistribution[i_d].distribution_1).required_subparticle, all_dddistribution[i_d].distribution_1.start, 0, all_dddistribution[i_d].distribution_1.n_bins * all_dddistribution[i_d].distribution_2.n_bins, all_dddistribution[i_d].distribution_1.step, "", all_dddistribution[i_d].distribution_1.type_binning, 0, 3);
      /*
      temp.xdistribution_name = all_dddistribution[i_d].name;
      temp.n_bins = all_dddistribution[i_d].n_bins;
      */
      extended_distribution.push_back(temp);
      mapping_dddistribution.push_back(i_d);

    }
  }

  cout << "extended_distribution.size() = " << extended_distribution.size() << endl;
  for (int d = 0; d < extended_distribution.size(); d++){cout << "extended_distribution[" << d << "] = " << endl << extended_distribution[d] << endl;}

  //  vector<double> fakeasymfactor(extended_distribution.size());
  fakeasymfactor.resize(extended_distribution.size(), 1.);
  for (int i_d = 0; i_d < extended_distribution.size(); i_d++){
    // everything below this line is repeated for all (extended) distributions !!!
    logger << LOG_DEBUG << "extended_distribution[" << i_d << "] = " << extended_distribution[i_d].xdistribution_name << endl;
    int ddd = mapping_dddistribution[i_d];
    //      if (i_d >= all_distribution.size()){ddd = i_d - all_distribution.size();}


    if (i_d >= all_distribution.size()){
      if (extended_distribution[i_d].xdistribution_type == "asym1"){fakeasymfactor[ddd] = (all_dddistribution[ddd].distribution_1).step;}
      else if (extended_distribution[i_d].xdistribution_type == "asym2"){fakeasymfactor[ddd] = (all_dddistribution[ddd].distribution_2).step;}
      if (extended_distribution[i_d].xdistribution_type == "sumasym1"){fakeasymfactor[ddd] = (all_dddistribution[ddd].distribution_1).step;}
      else if (extended_distribution[i_d].xdistribution_type == "sumasym2"){fakeasymfactor[ddd] = (all_dddistribution[ddd].distribution_2).step;}
    }
    else {fakeasymfactor[ddd] = extended_distribution[i_d].step;}
  }
}



void observable_set::get_userinput_extravalue_from_readin(vector<string> & user_variable, vector<string> & user_variable_additional, vector<string> & user_value, vector<string> & readin){
  Logger logger("get_userinput_extravalue_from_readin");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  string readindata;
  for (int i = 0; i < readin.size(); i++){
    readindata = readin[i][0];
    if (readindata != "/" && readindata != "#" && readindata != "%"){
      int start = 0;
      user_variable.push_back("");
      user_variable_additional.push_back("");
      user_value.push_back("");
      for (int j = 0; j < readin[i].size(); j++){
	if (start == 0 || start == 1){
	  if (((readin[i][j] == ' ') || (readin[i][j] == char(9))) && start == 0){}
	  else if ((readin[i][j] != ' ') && (readin[i][j] != char(9))){
	    user_variable[user_variable.size() - 1].push_back(readin[i][j]);
	    if (start != 1){start = 1;}
	  }
	  else {start++;}
	}
	else if (start == 2){
	  if (readin[i][j] == '='){start = 5;}
	  else if ((readin[i][j] == ' ') || (readin[i][j] == char(9))){}
	  else if (start == 2){
	    start++;
	    j--;
	  }
	  else {
	    logger << LOG_ERROR << "Incorrect input in line " << i << endl;
	    user_variable.erase(user_variable.end(), user_variable.end());
	    user_variable_additional.erase(user_variable_additional.end(), user_variable_additional.end());
	    user_value.erase(user_value.end(), user_value.end());
	    break;
	  }
	}
	else if (start == 3){
	  if ((readin[i][j] != '=')){
	    user_variable_additional[user_variable_additional.size() - 1].push_back(readin[i][j]);
	  }
	  else {start++; j--;} // should be the same as shifting (start == 4) here !!!
	}
	else if (start == 4){
	  // additional: should be allowed to contain ' ' !!!
	  if (readin[i][j] == '='){
	    start = 5;
	    logger << LOG_DEBUG << "before: user_variable_aditional: ---" << user_variable_additional[user_variable_additional.size() - 1] << "---" << endl;
	    for (int i_s = user_variable_additional[user_variable_additional.size() - 1].size() - 1; i_s > 0; i_s--){
	      if (user_variable_additional[user_variable_additional.size() - 1][i_s] == ' ' ||
		  user_variable_additional[user_variable_additional.size() - 1][i_s] == char(9)){
		user_variable_additional[user_variable_additional.size() - 1].erase(user_variable_additional[user_variable_additional.size() - 1].end() - 1, user_variable_additional[user_variable_additional.size() - 1].end());
	      }
	      else {break;}
	    }
	    logger << LOG_DEBUG << "after:  user_variable_aditional: ---" << user_variable_additional[user_variable_additional.size() - 1] << "---" << endl;
	  }
	  //	  else if ((readin[i][j] == ' ') || (readin[i][j] == char(9))){}
	  else {
	    logger << LOG_ERROR << "Incorrect input in line " << i << endl;
	    user_variable.erase(user_variable.end(), user_variable.end());
	    user_variable_additional.erase(user_variable_additional.end(), user_variable_additional.end());
	    user_value.erase(user_value.end(), user_value.end());
	    break;
	  }
	}
	else if (start == 5 || start == 6){
	  if (((readin[i][j] == ' ') || (readin[i][j] == char(9))) && start == 5){}
	  else if (readindata != "/" && readindata != "#" && readindata != "%"){
	    //	  else if ((readin[i][j] != ' ') && (readin[i][j] != char(9))){
	    user_value[user_value.size() - 1].push_back(readin[i][j]);
	    if (start != 6){start = 6;}
	  }
	  else {start++;}
	}
	else {break;}
      }
      logger << LOG_DEBUG << "i = " << setw(3) << i << "   ---" << user_value[user_value.size() - 1] << "---" << endl;
      for (int i_s = user_value[user_value.size() - 1].size() - 1; i_s > 0; i_s--){
	if (user_value[user_value.size() - 1][i_s] == ' ' ||
	    user_value[user_value.size() - 1][i_s] == char(9)){
	  user_value[user_value.size() - 1].erase(user_value[user_value.size() - 1].end() - 1, user_value[user_value.size() - 1].end());
	}
	else {break;}
      }
      logger << LOG_DEBUG << "i = " << setw(3) << i << "   ---" << user_value[user_value.size() - 1] << "---" << endl;
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}



void observable_set::distribution_back_to_back_configuration(int i_a, int k_g){
  Logger logger("observable_set::distribution_back_to_back_configuration");
  logger << LOG_DEBUG << "started" << endl;

  if (abs(particle_event[k_g][i_a][0].pT - particle_event[k_g][i_a][1].pT) / (particle_event[k_g][i_a][0].pT + particle_event[k_g][i_a][1].pT) < 1.e-12){
    if (particle_event[k_g][0].size() > 1){
      if (particle_event[k_g][i_a][0].momentum.x1() > 0. && particle_event[k_g][0][0].momentum.x1() < 0.){
	swap(particle_event[k_g][i_a][0], particle_event[k_g][i_a][1]);
	logger << LOG_DEBUG_VERBOSE << "Momenta have been swapped in order to match real phase space." << endl;
      } 
    }
  }

  logger << LOG_DEBUG << "finished" << endl;
}
