#!/usr/bin/python
import numpy as np
from argparse import ArgumentParser
import re
from matched_module import do_matching

def main():
    """
    """

    # read arguments from command line
    parser = ArgumentParser(description='Inputs')
    parser.add_argument('-f', '--fo_file', dest="fo_file",
                        help="fixed order file",
                        metavar="FOINFILE",
                        required=True)
    parser.add_argument('-r', '--res_file', dest="res_file",
                        help="resummation file",
                        metavar="RESINFILE",
                        required=True)
    parser.add_argument('-a', '--accuracy', dest="res_accuracy",
                        help="resummation accuracy", type=int,
                        metavar="RESUMMED ACCURACY",
                        required=True)
    parser.add_argument('-m', '--matched_order', dest="matched_order",
                        help="resummation file", type=int,
                        metavar="MATCHED ORDER",
                        required=True)
    parser.add_argument('-o', '--output', dest="output_file",
                        help="output file",
                        metavar="OFILE",
                        required=True)
    parser.add_argument('-b', '--bin', dest="bin_to_remove",
                        help="removing bin in fixed order files if they do not match",
                        metavar="BIN", type=int, default=0 )
    parser.add_argument('-l', '--lines', dest="lines_to_print",
                        help="number of lines at the top of the matched files (for backward-compatibility)",
                        metavar="LINES", type=int, default=0 )
    parser.add_argument('-c', '--columns', dest="extra_columns",
                        help="number of extra dummy columns of the matched files (for backward-compatibility)",
                        metavar="LINES", type=int, default=0 )
    args = parser.parse_args()

    # multiplicative scheme
    scheme = "mult"
    differential = True
    normalise = True

    # parse information from fixed-order file
    sigma = []
    filefo = open(args.fo_file, 'rb')
    while True:
        line = filefo.readline()
        if line.startswith("xsct_lo"):
            sigma.append(float(str.split(line)[2]))
        if line.startswith("delta_xsct_nlo"):
            sigma.append(float(str.split(line)[2]))
        if line.startswith("delta_xsct_nnlo"):
            sigma.append(float(str.split(line)[2]))
        if line.startswith("delta_xsct_nnnlo"):
            sigma.append(float(str.split(line)[2]))
        if not line:
            break

    # read sigmabar, resummed result and expansion
    v, sigmabar_1, sigmabar_2, sigmabar_3 = np.loadtxt(args.fo_file,unpack=True,skiprows=23)

    # remove last element as the number of bin changed
    if(args.bin_to_remove >0):
        print("Removing last bin of the fixed order files.")
        v = v[:-args.bin_to_remove]
        sigmabar_1 = sigmabar_1[:-args.bin_to_remove]
        sigmabar_2 = sigmabar_2[:-args.bin_to_remove]
        sigmabar_3 = sigmabar_3[:-args.bin_to_remove]

    # now parse information from resummed file
    fileres = open(args.res_file, 'rb')
    lumi_asy = []
    while True:
        line = fileres.readline()
        if line.startswith("# exp0="):
            lumi_asy.append(float(str.split(line)[2]))
        if line.startswith("# lumi_asympt0="):
            lumi_asy.append(float(str.split(line)[2]))
        if line.startswith("# lumi_asympt1="):
            lumi_asy.append(float(str.split(line)[2]))
        if line.startswith("# lumi_asympt2="):
            lumi_asy.append(float(str.split(line)[2]))
        if not line:
            break

    vres, dummy, resummed, dummy, exp, res_expanded_1, res_expanded_2, res_expanded_3 = np.loadtxt(args.res_file,unpack=True,skiprows=26)

    #determine number of bins
    nbins = len(v)
    if(nbins != len(vres)):
        print("Incompatible fixed order and resummed files. Please check the number of bins")
        print("Number of bins fixed order file: "+str(nbins))
        print("Number of bins resummed file: "+str(len(vres)))
        quit()

    #set resummation order and fixed order
    fo = args.matched_order
    ro = args.res_accuracy

    #compute matched result
    matched, dummy, fixed_order, dummy = do_matching(scheme, nbins, fo, ro, sigma, sigma, lumi_asy, lumi_asy, sigmabar_1, sigmabar_1, sigmabar_2, sigmabar_2, sigmabar_3, sigmabar_3, resummed, resummed, res_expanded_1, res_expanded_1, res_expanded_2, res_expanded_2, res_expanded_3, res_expanded_3, normalise)

    mat_result = []
    v_result = []
    res_result = []
    fo_result = []
    exp_result = []
    if(differential):
        v_result.append(v[0]/2)
        mat_result.append( (matched[0])/(v[0]) )
        res_result.append( (resummed[0])/(v[0]) )
        fo_result.append( 0 )
        exp_result.append( 0 )
        for i in range(len(matched)-1):
            v_result.append((v[i+1]+v[i])/2)
            mat_result.append( (matched[i+1]-matched[i])/(v[i+1]-v[i]) )
            res_result.append( (resummed[i+1]-resummed[i])/(v[i+1]-v[i]) )
            fo_result.append( (fixed_order[i+1]-fixed_order[i])/(v[i+1]-v[i]) )
            exp_result.append( (exp[i+1]-exp[i])/(v[i+1]-v[i]) )
    else:
        for i in range(len(matched)):
            v_result.append(v[i])
            mat_result.append( matched[i] )
            res_result.append( resummed[i] )
            fo_result.append( fixed_order[i] )
            exp_result.append( exp[i] )

    sigmatot = sum(sigma[0:fo])
    if(normalise):
        for i, res in enumerate(res_result):
            res_result[i] = res/sigmatot
        for i, exp in enumerate(exp_result):
            exp_result[i] = exp/sigmatot

    file_output = open(args.output_file, "w")

    #print matched result

    # additional lines at the top of the file
    if(args.lines_to_print>0):
        for i in range(0,args.lines_to_print):
            file_output.write("#\n")

    for i in range(0,len(mat_result)):
        file_output.write(str(v_result[i]) + " " + str(mat_result[i]) +
                          " " + str(res_result[i]) + " " + str(fo_result[i]) +
                          " " + str(exp_result[i]))
        for j in range (0,args.extra_columns):
            file_output.write(" " + "0.0")
        file_output.write("\n")

if __name__ == '__main__':
    main()
