#!/usr/bin/python

""" matrix_interface wrapper using ctypes. """
import ctypes

# find and load dynamic library (OSX or linux)
ctypes.cdll.LoadLibrary("/libchaplin.dylib")
ctypes.cdll.LoadLibrary("/usr/local/Cellar/lhapdf/6.2.3-clang/lib/liblhapdf.dylib")
MYLIB = ctypes.cdll.LoadLibrary("/Users/rottoli/codes/mycodes/mc-resummation-cern/RadISH/libradish.dylib")

# set the argument type
MYLIB.do_matching.argtypes = [ctypes.c_char_p,
                              ctypes.c_int, ctypes.c_int, ctypes.c_int,
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double), 
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.c_bool, ctypes.c_bool, ctypes.c_char_p, ctypes.c_double]

MYLIB.do_matching_matrix.argtypes = [ctypes.c_char_p,
                              ctypes.c_int, ctypes.c_int, ctypes.c_int,
                              ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double),
                              ctypes.c_bool]
# set the return type
MYLIB.do_matching.restype = ctypes.c_void_p

MYLIB.do_matching_matrix.restype = ctypes.c_void_p

def do_matching(scheme, nbins, ford, rord, sigma_in, sigma_err_in, lumi_asy_in,
                lumi_asy_err_in, v_in, sigmabar_1_in, sigmabar_1_err_in, sigmabar_2_in,
                sigmabar_2_err_in, sigmabar_3_in, sigmabar_3_err_in, resummed_in,
                resummed_err_in, res_expanded_1_in, res_expanded_1_err_in, res_expanded_2_in,
                res_expanded_2_err_in, res_expanded_3_in, res_expanded_3_err_in,
                normalised_in=False, damping_in=False, observable='ptB', scale_in=0.):
    """Python wrapper for matching routine 'do_matching' in radish.h.

    Args:
        scheme (string): matching scheme; can be either additive ('add') or multiplicative ('mult')
        nbins     (int): number of bins
        ford        (int): fixed order (0: LO, 1: NLO, 2: NNLO, 3: N3LO)
        rord        (int): resum order (1: NLL, 2: NNLL, 3: N3LL)
        sigma[4]  (double)     : fixed order cross sections (LO,deltaNLO,deltaNNLO,deltaN3LO)
        sigma_err[4] (double)  : corresponding error
        lumi_asy[4]  (double)   : expansion and asymptotic luminosities
            (res_0, lumi_NLL, lumi_NNLL, lumi_N3LL)
        lumi_asy_error[4] (double) : corresponding error
        v_in[nbins] (double): value of the observable
        sigmabar_1_in[nbins] (double): values of as^0 contribution to sigmabar
        sigmabar_1_err_in[nbins] (double): corresponding error
        sigmabar_2_in[nbins] (double): values of as^1 contribution to sigmabar
        sigmabar_2_err_in[nbins] (double): corresponding error
        sigmabar_3_in[nbins] (double): values of as^2 contribution to sigmabar
        sigmabar_3_err_in[nbins] (double): corresponding error
        resummed_in[nbins] (double): resummed contribution
        resummed_err_in[nbins] (double): corresponding error
        res_expanded_1_in[nbins] (double): order as^1 expansion
        res_expanded_1_in_err_in[nbins] (double): corresponding error
        res_expanded_2_in_in[nbins] (double): order as^2 expansion
        res_expanded_2_in_err_in[nbins] (double): corresponding error
        res_expanded_3_in_in[nbins] (double): order as^3 expansion
        res_expanded_3_in_err_in[nbins] (double): corresponding error
        normalised (bool): normalise to the total cross section (default false)
        damping (bool): whether to add a damping to the multiplicative scheme (default false)
        observable (string): needed if damping is set to true (default ptB)
        scale_in (double): scale where the damping start to have an effect (default 90.)

    Returns:
        double[nbins], double[nbins], double[nbins], double[nbins]:
        matched result, matched result error, fixed order result, fixed order result error

    """

    if(ford < 0 or ford > 3):
        print("Fixed order must be between 0 and 3")
        quit()

    if(rord < 1 or rord > 3):
        print("Resummed order must be between 1 and 3")
        quit()

    #cast everything using ctypes to call the c++ function
    sigma = (ctypes.c_double * len(sigma_in))(*sigma_in)
    sigma_err = (ctypes.c_double * len(sigma_err_in))(*sigma_err_in)

    lumi_asy = (ctypes.c_double * len(lumi_asy_in))(*lumi_asy_in)
    lumi_asy_err = (ctypes.c_double * len(lumi_asy_err_in))(*lumi_asy_err_in)

    v = (ctypes.c_double * len(v_in))(*v_in)

    sigmabar_1 = (ctypes.c_double * len(sigmabar_1_in))(*sigmabar_1_in)
    sigmabar_2 = (ctypes.c_double * len(sigmabar_2_in))(*sigmabar_2_in)
    sigmabar_3 = (ctypes.c_double * len(sigmabar_3_in))(*sigmabar_3_in)

    sigmabar_1_err = (ctypes.c_double * len(sigmabar_1_err_in))(*sigmabar_1_err_in)
    sigmabar_2_err = (ctypes.c_double * len(sigmabar_2_err_in))(*sigmabar_2_err_in)
    sigmabar_3_err = (ctypes.c_double * len(sigmabar_3_err_in))(*sigmabar_3_err_in)

    resummed = (ctypes.c_double * len(resummed_in))(*resummed_in)
    resummed_err = (ctypes.c_double * len(resummed_err_in))(*resummed_err_in)

    res_expanded_1 = (ctypes.c_double * len(res_expanded_1_in))(*res_expanded_1_in)
    res_expanded_2 = (ctypes.c_double * len(res_expanded_2_in))(*res_expanded_2_in)
    res_expanded_3 = (ctypes.c_double * len(res_expanded_3_in))(*res_expanded_3_in)

    res_expanded_1_err = (ctypes.c_double * len(res_expanded_1_err_in))(*res_expanded_1_err_in)
    res_expanded_2_err = (ctypes.c_double * len(res_expanded_2_err_in))(*res_expanded_2_err_in)
    res_expanded_3_err = (ctypes.c_double * len(res_expanded_3_err_in))(*res_expanded_3_err_in)

    normalised = ctypes.c_bool(normalised_in)

    #pointer to matched and fixed_order result
    matched = []
    matched_cumulative = (ctypes.c_double * len(resummed))(*matched)
    fixed_order = []
    fixed_order_result = (ctypes.c_double * len(resummed))(*fixed_order)
    expansion = []
    expansion_result = (ctypes.c_double * len(resummed))(*expansion)

    matched_cumulative_error = []
    fixed_order_result_error = []
    expansion_result_error   = []
    for _ in range(nbins):
        matched_cumulative_error.append(0)
        fixed_order_result_error.append(0)
        expansion_result_error.append(0)

    damping = ctypes.c_bool(damping_in)
    scale = ctypes.c_double(scale_in)

    #call the c++ function
    MYLIB.do_matching(scheme.encode('utf-8'), nbins, ford, rord, sigma, sigma_err,
                      lumi_asy, lumi_asy_err, v, sigmabar_1, sigmabar_1_err, sigmabar_2,
                      sigmabar_2_err, sigmabar_3, sigmabar_3_err, resummed,
                      resummed_err, res_expanded_1, res_expanded_1_err, res_expanded_2,
                      res_expanded_2_err, res_expanded_3, res_expanded_3_err,
                      matched_cumulative, fixed_order_result, expansion_result, normalised, damping, observable.encode('utf-8'), scale)

    #return array of matched results and fixed order result
    return matched_cumulative, matched_cumulative_error, fixed_order_result, fixed_order_result_error, expansion_result, expansion_result_error

def do_matching_matrix(scheme, nbins, ford, rord, sigmatot_in, sigma_err_in, lumi_asy_in,
                lumi_asy_err_in, sigma_1_in, sigma_1_err_in, sigma_2_in,
                sigma_2_err_in, sigma_3_in, sigma_3_err_in, resummed_in,
                resummed_err_in, res_expanded_1_in, res_expanded_1_err_in, res_expanded_2_in,
                res_expanded_2_err_in, res_expanded_3_in, res_expanded_3_err_in,
                normalised_in=False):
    """Python wrapper for matching routine 'do_matching' in radish.h.

    Args:
        scheme (string): matching scheme; can be either additive ('add') or multiplicative ('mult')
        nbins     (int): number of bins
        ford        (int): fixed order (0: LO, 1: NLO, 2: NNLO, 3: N3LO)
        rord        (int): resum order (1: NLL, 2: NNLL, 3: N3LL)
        sigma[4]  (double)     : fixed order cross sections (LO,deltaNLO,deltaNNLO,deltaN3LO)
        sigma_err[4] (double)  : corresponding error
        lumi_asy[4]  (double)   : expansion and asymptotic luminosities
            (res_0, lumi_NLL, lumi_NNLL, lumi_N3LL)
        lumi_asy_error[4] (double) : corresponding error
        sigmabar_1_in[nbins] (double): values of as^0 contribution to sigmabar
        sigmabar_1_err_in[nbins] (double): corresponding error
        sigmabar_2_in[nbins] (double): values of as^1 contribution to sigmabar
        sigmabar_2_err_in[nbins] (double): corresponding error
        sigmabar_3_in[nbins] (double): values of as^2 contribution to sigmabar
        sigmabar_3_err_in[nbins] (double): corresponding error
        resummed_in[nbins] (double): resummed contribution
        resummed_err_in[nbins] (double): corresponding error
        res_expanded_1_in[nbins] (double): order as^1 expansion
        res_expanded_1_in_err_in[nbins] (double): corresponding error
        res_expanded_2_in_in[nbins] (double): order as^2 expansion
        res_expanded_2_in_err_in[nbins] (double): corresponding error
        res_expanded_3_in_in[nbins] (double): order as^3 expansion
        res_expanded_3_in_err_in[nbins] (double): corresponding error
        normalised (bool): normalise to the total cross section (default false)

    Returns:
        double[nbins], double[nbins], double[nbins], double[nbins]:
        matched result, matched result error, fixed order result, fixed order result error

    """

    if(ford < 0 or ford > 3):
        print("Fixed order must be between 0 and 3")
        quit()

    if(rord < 1 or rord > 3):
        print("Resummed order must be between 1 and 3")
        quit()

    #cast everything using ctypes to call the c++ function
    sigmatot = (ctypes.c_double * len(sigmatot_in))(*sigmatot_in)
    sigmatot_err = (ctypes.c_double * len(sigma_err_in))(*sigma_err_in)

    lumi_asy = (ctypes.c_double * len(lumi_asy_in))(*lumi_asy_in)
    lumi_asy_err = (ctypes.c_double * len(lumi_asy_err_in))(*lumi_asy_err_in)

    sigma_1 = (ctypes.c_double * len(sigma_1_in))(*sigma_1_in)
    sigma_2 = (ctypes.c_double * len(sigma_2_in))(*sigma_2_in)
    sigma_3 = (ctypes.c_double * len(sigma_3_in))(*sigma_3_in)

    sigma_1_err = (ctypes.c_double * len(sigma_1_err_in))(*sigma_1_err_in)
    sigma_2_err = (ctypes.c_double * len(sigma_2_err_in))(*sigma_2_err_in)
    sigma_3_err = (ctypes.c_double * len(sigma_3_err_in))(*sigma_3_err_in)

    resummed = (ctypes.c_double * len(resummed_in))(*resummed_in)
    resummed_err = (ctypes.c_double * len(resummed_err_in))(*resummed_err_in)

    res_expanded_1 = (ctypes.c_double * len(res_expanded_1_in))(*res_expanded_1_in)
    res_expanded_2 = (ctypes.c_double * len(res_expanded_2_in))(*res_expanded_2_in)
    res_expanded_3 = (ctypes.c_double * len(res_expanded_3_in))(*res_expanded_3_in)

    res_expanded_1_err = (ctypes.c_double * len(res_expanded_1_err_in))(*res_expanded_1_err_in)
    res_expanded_2_err = (ctypes.c_double * len(res_expanded_2_err_in))(*res_expanded_2_err_in)
    res_expanded_3_err = (ctypes.c_double * len(res_expanded_3_err_in))(*res_expanded_3_err_in)

    normalised = ctypes.c_bool(normalised_in)

    #pointer to matched and fixed_order result
    matched = []
    matched_cumulative = (ctypes.c_double * len(resummed))(*matched)
    fixed_order = []
    fixed_order_result = (ctypes.c_double * len(resummed))(*fixed_order)
    expansion = []
    expansion_result = (ctypes.c_double * len(resummed))(*expansion)

    matched_cumulative_error = []
    fixed_order_result_error = []
    expansion_result_error   = []
    for _ in range(nbins):
        matched_cumulative_error.append(0)
        fixed_order_result_error.append(0)
        expansion_result_error.append(0)

    #call the c++ function
    MYLIB.do_matching_matrix(scheme.encode('utf-8'), nbins, ford, rord, sigmatot, sigmatot_err,
                      lumi_asy, lumi_asy_err, sigma_1, sigma_1_err, sigma_2,
                      sigma_2_err, sigma_3, sigma_3_err, resummed,
                      resummed_err, res_expanded_1, res_expanded_1_err, res_expanded_2,
                      res_expanded_2_err, res_expanded_3, res_expanded_3_err,
                      matched_cumulative, fixed_order_result, expansion_result, normalised)

    #return array of matched results and fixed order result
    return matched_cumulative, matched_cumulative_error, fixed_order_result, fixed_order_result_error, expansion_result, expansion_result_error
