module gen_radiation
   use types; use consts_dp
   use ew_parameters
   use shower_routines
   use histogram_routines
   use sub_defs_io
   use rad_tools
   use hoppet_v1
   use expansion
   use masters

   implicit none

   private :: event_gen
   private :: init_mcfm
   public :: gen_rad
   public :: gen_rad_inclusive
   public :: terminate_radish

   interface
      subroutine event_gen()
      end subroutine event_gen
   end interface

contains

!#######################################################################
   subroutine gen_rad(order, nev, ncall1, ncall2, gen_seed, usecuts)
      include 'mcfmparameters.f'
      include 'cuts.f'
      logical, intent(in) :: usecuts
      integer, intent(in) :: order, nev, ncall1, ncall2, gen_seed ! these are parameters for the Born generator
      integer :: i

      ! set up the run
      use_cuts = usecuts ! this is saved in cuts.f
      ! initialise shower
      call init_shower(order, nev)

      ! make input card
      call init_mcfm(ncall1, ncall2, gen_seed)

      ! generate events
      call event_gen()

      ! normalise histograms
      call normalise_histograms(born_itmx2, nev)

      ! call terminate_radish()

   end subroutine gen_rad

!#######################################################################
   subroutine gen_rad_inclusive(order, nev)
      integer, intent(in) :: order, nev
      integer :: i

      ! initialise shower
      call init_shower(order, nev)

      if ((cs%observable == 'ptj') .and. (cs%numerical_jetveto .eqv. .false.)) then
         call jetveto_inclusive()
      else
         call make_shower_inclusive()
      endif

      ! normalise histograms
      call normalise_histograms(1, nev)

      ! call terminate_radish()

   end subroutine gen_rad_inclusive

   ! routine to deallocate used memory
   subroutine terminate_radish() BIND(C)
      use iso_c_binding

      call Delete(pdf1e)
      call Delete(pdf2e)

   end subroutine terminate_radish

   !#######################################################################
   ! This routine passes the relevant parameters to the MCFM integrator
   subroutine init_mcfm(ncall1, ncall2, gen_seed)
      integer, intent(in) :: ncall1, ncall2, gen_seed
      include 'mcfmparameters.f'
      include 'cuts.f'
      born_proc = cs%proc
      if (cs%DY_fixedMY) then
         born_proc = 'DYMY'
         born_fixed_M = cs%fixed_M
         born_fixed_Y = cs%fixed_Y
      endif
      born_sqrts = cs%rts
      if (cs%collider == 'pp') then
         born_ih1 = 1
         born_ih2 = 1
      elseif (cs%collider == 'ppbar') then
         born_ih1 = 1
         born_ih2 = -1
      end if
      born_hmass = cs%M
      born_scale = cs%muR
      born_facscale = cs%muF
      born_itmx1 = 4
      born_ncall1 = ncall1
      born_itmx2 = 4
      born_ncall2 = ncall2
      born_origij = gen_seed
      born_ewscheme = +1
      born_GF_GeVm2 = GF_GeVm2
      born_aemmz = aemmz
      born_sinwsq = sinwsq
      born_MW = MW
      born_MZ = MZ
      born_Wwidth = Wwidth
      born_Zwidth = Zwidth
      born_Hwidth = Hwidth
      born_nf = nf_int
      born_mt = mt
      born_mb = mb
      born_mc = mc
      born_ms = ms
      born_Vud = Vud_in
      born_Vus = Vus_in
      born_Vub = Vub_in
      born_Vcd = Vcd_in
      born_Vcs = Vcs_in
      born_Vcb = Vcb_in
      if (.not. use_cuts) then
         if (cs%proc .eq. "H") then
            born_m_min = 1d0
            born_m_max = cs%rts
         else
            born_m_min = cs%m_min
            born_m_max = cs%m_max
         endif
         born_eta_min = 0d0
         born_eta_max = +100d0
      else
         born_m_min = cs%m_min
         born_m_max = cs%m_max
         born_eta_min = cs%eta_min
         born_eta_max = cs%eta_max
      end if
      mcfm_dryrun = cs%dryrun
      hdecay_mode = cs%hdecay_mode
   end subroutine init_mcfm

end module gen_radiation
